/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/psp/texture.c: Texture manipulation functionality for the PSP.
 */

/*
 * Notes about PSP hardware limitations:
 *
 * The PSP's graphics hardware has a maximum texture size of 512x512
 * pixels.  Currently, SIL does not restrict the size of textures created
 * with the texture_create(), texture_parse(), and related functions; it
 * will happily let you create (for example) a 2048x2048 texture, though
 * attempting to render using that texture will have undefined results.
 *
 * However, SIL is capable of rendering with oversize textures under
 * limited conditions:
 *
 * - When rendering a QUADS primitive, if the height of the current texture
 *   is greater than 512, SIL will check the V coordinate of the first
 *   vertex of each quad.  Treating the texture as a vertically stacked set
 *   of 512-high subtextures, if the V coordinate of the new triangle is in
 *   a different subtexture than the V coordinate of the last triangle, SIL
 *   will adjust the texture pointers and restart the primitive so that the
 *   correct portion of the texture is drawn.  This adjustment is only
 *   performed the first time the primitive is drawn, and only if an
 *   oversize texture is applied at that time.  For this to work correctly,
 *   each quad must be completely contained in a single subtexture, and the
 *   primitive must always be drawn using the same texture (or another
 *   texture with the same size).
 *
 *   This feature is primarily intended to support bitmap fonts which have
 *   too many characters to fit in a 512x512 texture, such as Japanese fonts.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/endian.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/psp/ge-util.h"
#include "src/sysdep/psp/internal.h"
#include "src/sysdep/psp/texture.h"
#include "src/texture.h"
#include "src/utility/tex-file.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Currently-applied texture.  We track this to avoid unnecessary clearing
 * of the GE's texture cache. */
static SysTexture *current_texture;

/* List of textures whose deletion has been deferred (see
 * sys_texture_destroy() for details). */
static SysTexture *deferred_destroy_list;

/*-----------------------------------------------------------------------*/

/* Static palette for alpha-only images. */
static const ALIGNED(64) uint32_t alpha_palette[256] = {
    0x00FFFFFF, 0x01FFFFFF, 0x02FFFFFF, 0x03FFFFFF,
    0x04FFFFFF, 0x05FFFFFF, 0x06FFFFFF, 0x07FFFFFF,
    0x08FFFFFF, 0x09FFFFFF, 0x0AFFFFFF, 0x0BFFFFFF,
    0x0CFFFFFF, 0x0DFFFFFF, 0x0EFFFFFF, 0x0FFFFFFF,
    0x10FFFFFF, 0x11FFFFFF, 0x12FFFFFF, 0x13FFFFFF,
    0x14FFFFFF, 0x15FFFFFF, 0x16FFFFFF, 0x17FFFFFF,
    0x18FFFFFF, 0x19FFFFFF, 0x1AFFFFFF, 0x1BFFFFFF,
    0x1CFFFFFF, 0x1DFFFFFF, 0x1EFFFFFF, 0x1FFFFFFF,
    0x20FFFFFF, 0x21FFFFFF, 0x22FFFFFF, 0x23FFFFFF,
    0x24FFFFFF, 0x25FFFFFF, 0x26FFFFFF, 0x27FFFFFF,
    0x28FFFFFF, 0x29FFFFFF, 0x2AFFFFFF, 0x2BFFFFFF,
    0x2CFFFFFF, 0x2DFFFFFF, 0x2EFFFFFF, 0x2FFFFFFF,
    0x30FFFFFF, 0x31FFFFFF, 0x32FFFFFF, 0x33FFFFFF,
    0x34FFFFFF, 0x35FFFFFF, 0x36FFFFFF, 0x37FFFFFF,
    0x38FFFFFF, 0x39FFFFFF, 0x3AFFFFFF, 0x3BFFFFFF,
    0x3CFFFFFF, 0x3DFFFFFF, 0x3EFFFFFF, 0x3FFFFFFF,
    0x40FFFFFF, 0x41FFFFFF, 0x42FFFFFF, 0x43FFFFFF,
    0x44FFFFFF, 0x45FFFFFF, 0x46FFFFFF, 0x47FFFFFF,
    0x48FFFFFF, 0x49FFFFFF, 0x4AFFFFFF, 0x4BFFFFFF,
    0x4CFFFFFF, 0x4DFFFFFF, 0x4EFFFFFF, 0x4FFFFFFF,
    0x50FFFFFF, 0x51FFFFFF, 0x52FFFFFF, 0x53FFFFFF,
    0x54FFFFFF, 0x55FFFFFF, 0x56FFFFFF, 0x57FFFFFF,
    0x58FFFFFF, 0x59FFFFFF, 0x5AFFFFFF, 0x5BFFFFFF,
    0x5CFFFFFF, 0x5DFFFFFF, 0x5EFFFFFF, 0x5FFFFFFF,
    0x60FFFFFF, 0x61FFFFFF, 0x62FFFFFF, 0x63FFFFFF,
    0x64FFFFFF, 0x65FFFFFF, 0x66FFFFFF, 0x67FFFFFF,
    0x68FFFFFF, 0x69FFFFFF, 0x6AFFFFFF, 0x6BFFFFFF,
    0x6CFFFFFF, 0x6DFFFFFF, 0x6EFFFFFF, 0x6FFFFFFF,
    0x70FFFFFF, 0x71FFFFFF, 0x72FFFFFF, 0x73FFFFFF,
    0x74FFFFFF, 0x75FFFFFF, 0x76FFFFFF, 0x77FFFFFF,
    0x78FFFFFF, 0x79FFFFFF, 0x7AFFFFFF, 0x7BFFFFFF,
    0x7CFFFFFF, 0x7DFFFFFF, 0x7EFFFFFF, 0x7FFFFFFF,
    0x80FFFFFF, 0x81FFFFFF, 0x82FFFFFF, 0x83FFFFFF,
    0x84FFFFFF, 0x85FFFFFF, 0x86FFFFFF, 0x87FFFFFF,
    0x88FFFFFF, 0x89FFFFFF, 0x8AFFFFFF, 0x8BFFFFFF,
    0x8CFFFFFF, 0x8DFFFFFF, 0x8EFFFFFF, 0x8FFFFFFF,
    0x90FFFFFF, 0x91FFFFFF, 0x92FFFFFF, 0x93FFFFFF,
    0x94FFFFFF, 0x95FFFFFF, 0x96FFFFFF, 0x97FFFFFF,
    0x98FFFFFF, 0x99FFFFFF, 0x9AFFFFFF, 0x9BFFFFFF,
    0x9CFFFFFF, 0x9DFFFFFF, 0x9EFFFFFF, 0x9FFFFFFF,
    0xA0FFFFFF, 0xA1FFFFFF, 0xA2FFFFFF, 0xA3FFFFFF,
    0xA4FFFFFF, 0xA5FFFFFF, 0xA6FFFFFF, 0xA7FFFFFF,
    0xA8FFFFFF, 0xA9FFFFFF, 0xAAFFFFFF, 0xABFFFFFF,
    0xACFFFFFF, 0xADFFFFFF, 0xAEFFFFFF, 0xAFFFFFFF,
    0xB0FFFFFF, 0xB1FFFFFF, 0xB2FFFFFF, 0xB3FFFFFF,
    0xB4FFFFFF, 0xB5FFFFFF, 0xB6FFFFFF, 0xB7FFFFFF,
    0xB8FFFFFF, 0xB9FFFFFF, 0xBAFFFFFF, 0xBBFFFFFF,
    0xBCFFFFFF, 0xBDFFFFFF, 0xBEFFFFFF, 0xBFFFFFFF,
    0xC0FFFFFF, 0xC1FFFFFF, 0xC2FFFFFF, 0xC3FFFFFF,
    0xC4FFFFFF, 0xC5FFFFFF, 0xC6FFFFFF, 0xC7FFFFFF,
    0xC8FFFFFF, 0xC9FFFFFF, 0xCAFFFFFF, 0xCBFFFFFF,
    0xCCFFFFFF, 0xCDFFFFFF, 0xCEFFFFFF, 0xCFFFFFFF,
    0xD0FFFFFF, 0xD1FFFFFF, 0xD2FFFFFF, 0xD3FFFFFF,
    0xD4FFFFFF, 0xD5FFFFFF, 0xD6FFFFFF, 0xD7FFFFFF,
    0xD8FFFFFF, 0xD9FFFFFF, 0xDAFFFFFF, 0xDBFFFFFF,
    0xDCFFFFFF, 0xDDFFFFFF, 0xDEFFFFFF, 0xDFFFFFFF,
    0xE0FFFFFF, 0xE1FFFFFF, 0xE2FFFFFF, 0xE3FFFFFF,
    0xE4FFFFFF, 0xE5FFFFFF, 0xE6FFFFFF, 0xE7FFFFFF,
    0xE8FFFFFF, 0xE9FFFFFF, 0xEAFFFFFF, 0xEBFFFFFF,
    0xECFFFFFF, 0xEDFFFFFF, 0xEEFFFFFF, 0xEFFFFFFF,
    0xF0FFFFFF, 0xF1FFFFFF, 0xF2FFFFFF, 0xF3FFFFFF,
    0xF4FFFFFF, 0xF5FFFFFF, 0xF6FFFFFF, 0xF7FFFFFF,
    0xF8FFFFFF, 0xF9FFFFFF, 0xFAFFFFFF, 0xFBFFFFFF,
    0xFCFFFFFF, 0xFDFFFFFF, 0xFEFFFFFF, 0xFFFFFFFF,
};

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * texture_to_RGBA:  Copy texture pixel data into an RGBA buffer.
 *
 * [Parameters]
 *     texture: Texture to copy from.
 *     dest: Destination buffer.
 */
static void texture_to_RGBA(SysTexture *texture, uint32_t *dest);

/**
 * pixel_16bpp_to_rgba:  Convert a 16-bit pixel value to a 32-bit RGBA value
 * (a 32-bit value which will be in RGBA byte order when stored in memory).
 *
 * [Parameters]
 *     format: Input pixel format (one of GE_TEXFMT_5650, GE_TEXFMT_5551,
 *         or GE_TEXFMT_4444).
 *     pixel: Input pixel value.
 * [Return value]
 *     Equivalent 32-bit pixel value.
 */
static inline uint32_t pixel_16bpp_to_RGBA(
    const int format, const uint16_t pixel);

/**
 * RGBA_to_texture:  Copy RGBA pixel data into a texture.
 *
 * [Parameters]
 *     src: Source buffer.
 *     src_stride: Source buffer line length, in pixels.
 *     width, height: Size of region to copy.
 *     has_alpha: True if the source alpha byte is valid, false if not.
 *     texture: Destination texture.
 */
static void RGBA_to_texture(const uint32_t *src, int src_stride,
                            int width, int height, int has_alpha,
                            SysTexture *texture);

/**
 * fb_to_texture:  Copy framebuffer data into a texture.
 *
 * [Parameters]
 *     x0, y0: Coordinates of first pixel to copy.
 *     width, height: Size of region to copy.
 *     texture: Destination texture.
 */
static void fb_to_texture(int x0, int y0, int width, int height,
                          SysTexture *texture);

/*************************************************************************/
/*************** Interface: Texture creation and deletion ****************/
/*************************************************************************/

SysTexture *sys_texture_create(
    int width, int height, TextureFormat data_format, int num_levels,
    void *data, int stride, const int32_t *level_offsets,
    const int32_t *level_sizes, UNUSED int mipmaps, int mem_flags, int reuse)
{
    /* First off, drop any excess image levels we can't use. */

    if (num_levels > 8) {
        DLOG("Warning: dropping unusable mipmap levels (%d > 8)", num_levels);
        num_levels = 8;
    }

    /* Check the data format and make sure it's something we can handle. */

    int format, bpp, is_alpha, is_aligned = 0, is_swizzled = 0;
    switch (data_format) {
      case TEX_FORMAT_PSP_RGBA8888_SWIZZLED:
        is_swizzled = 1;
        /* Fall through. */
      case TEX_FORMAT_PSP_RGBA8888:
        is_aligned = 1;
        /* Fall through. */
      case TEX_FORMAT_RGBA8888:
        format = GE_TEXFMT_8888;
        bpp = 32;
        is_alpha = 0;
        break;

      case TEX_FORMAT_PSP_RGB565_SWIZZLED:
        is_swizzled = 1;
        /* Fall through. */
      case TEX_FORMAT_PSP_RGB565:
        is_aligned = 1;
        /* Fall through. */
      case TEX_FORMAT_RGB565:
        format = GE_TEXFMT_5650;
        bpp = 16;
        is_alpha = 0;
        break;

      case TEX_FORMAT_PSP_RGBA5551_SWIZZLED:
        is_swizzled = 1;
        /* Fall through. */
      case TEX_FORMAT_PSP_RGBA5551:
        is_aligned = 1;
        /* Fall through. */
      case TEX_FORMAT_RGBA5551:
        format = GE_TEXFMT_5551;
        bpp = 16;
        is_alpha = 0;
        break;

      case TEX_FORMAT_PSP_RGBA4444_SWIZZLED:
        is_swizzled = 1;
        /* Fall through. */
      case TEX_FORMAT_PSP_RGBA4444:
        is_aligned = 1;
        /* Fall through. */
      case TEX_FORMAT_RGBA4444:
        format = GE_TEXFMT_4444;
        bpp = 16;
        is_alpha = 0;
        break;

      case TEX_FORMAT_PSP_A8_SWIZZLED:
        is_swizzled = 1;
        /* Fall through. */
      case TEX_FORMAT_PSP_A8:
        is_aligned = 1;
        /* Fall through. */
      case TEX_FORMAT_A8:
        format = GE_TEXFMT_T8;
        bpp = 8;
        is_alpha = 1;
        break;

      case TEX_FORMAT_PSP_PALETTE8_RGBA8888_SWIZZLED:
        is_swizzled = 1;
        /* Fall through. */
      case TEX_FORMAT_PSP_PALETTE8_RGBA8888:
        is_aligned = 1;
        /* Fall through. */
      case TEX_FORMAT_PALETTE8_RGBA8888:
        format = GE_TEXFMT_T8;
        bpp = 8;
        is_alpha = 0;
        break;

      default:
        DLOG("Pixel format %u unsupported", data_format);
        goto error_return;
    }

    /* If data reuse was requested, check whether we can in fact reuse the
     * data buffer. */

    int do_reuse = 0;
    if (num_levels > 0 && reuse) {
        if (!is_aligned && (stride % (128/bpp) != 0 || num_levels > 1)) {
            DLOG("Not reusing data: stride (%d) is not a multiple of 16 bytes"
                 " or texture has mipmaps", stride);
        } else if (level_offsets[0] < (int)sizeof(SysTexture)) {
            DLOG("Not reusing data: not enough space at beginning of buffer"
                 " (%zu, need %zu)", level_offsets[0], sizeof(SysTexture));
        } else {
            do_reuse = 1;
            for (int level = 0; level < num_levels; level++) {
                if (((uintptr_t)data + level_offsets[level]) % 64 != 0) {
                    DLOG("Not reusing data: level %d (%p) not 64-byte aligned",
                         level, (uint8_t *)data + level_offsets[level]);
                    do_reuse = 0;
                    break;
                }
            }
        }
    }

    /* Set up a SysTexture structure for the texture.  If we're not reusing
     * the input data buffer, allocate the SysTexture structure and texture
     * data as a single memory block for simplicity. */

    SysTexture *texture;
    uint8_t *texture_limit;  // For sceKernelDcacheWritebackRange().
    const uint32_t *palette = is_alpha ? alpha_palette : NULL;

    if (do_reuse) {

        mem_debug_set_info(data, MEM_INFO_TEXTURE);
        texture = (SysTexture *)data;
        texture_limit = NULL;  // Not used.
        texture->stride = stride;
        for (int level = 0; level < num_levels; level++) {
            texture->pixels[level] = (uint8_t *)data + level_offsets[level];
            sceKernelDcacheWritebackRange(texture->pixels[level],
                                          level_sizes[level]);
        }
        if (bpp == 8 && !is_alpha) {
            palette = (uint32_t *)(void *)(texture->pixels[0]);
            sceKernelDcacheWritebackRange(palette, 256*4);
            texture->pixels[0] += 256*4;
        }
        reuse = 0;  // Don't free the buffer on return.

    } else {  // !do_reuse

        uint32_t my_level_sizes[8];
        const int32_t struct_size = align_up(sizeof(SysTexture), 64);
        int32_t total_size = struct_size;
        if (bpp == 8 && !is_alpha) {
            total_size += 256*4;
        }
        int level_h = height;
        int level_s = align_up(width, 128/bpp);
        for (int level = 0; level < lbound(num_levels,1); level++) {
            const int level_size = (level_h * level_s * bpp) / 8;
            my_level_sizes[level] = align_up(level_size, 64);
            total_size += my_level_sizes[level];
            level_h = lbound(level_h/2, 1);
            level_s = align_up(lbound(level_s/2, 1), 128/bpp);
        }

        texture = debug_mem_alloc(total_size, 64, mem_flags,
                                  __FILE__, __LINE__, MEM_INFO_TEXTURE);
        if (UNLIKELY(!texture)) {
            DLOG("Failed to allocate %dx%d texture (%d levels, %zu bytes)",
                 width, height, lbound(num_levels,1), total_size);
            goto error_return;
        }
        texture_limit = (uint8_t *)texture + total_size;
        texture->stride = align_up(width, 128/bpp);
        uint8_t *pixels = (uint8_t *)texture + struct_size;
        if (bpp == 8 && !is_alpha) {
            /* Do the copy here so we don't have to de-const below. */
            if (num_levels > 0) {
                memcpy(pixels, (uint8_t *)data + level_offsets[0], 256*4);
                sceKernelDcacheWritebackRange(pixels, 256*4);
            } else {
                DLOG("Warning: indexed texture without data, assuming"
                     " alpha-only palette");
                memcpy(pixels, alpha_palette, 256*4);
            }
            /* Cast through void * to suppress cast-align warning. */
            palette = (const uint32_t *)(void *)pixels;
            pixels += 256*4;
        }
        for (int level = 0; level < lbound(num_levels,1); level++) {
            texture->pixels[level] = pixels;
            pixels += my_level_sizes[level];
        }

    }  // do_reuse

    texture->width     = width;
    texture->height    = height;
    texture->format    = format;
    texture->swizzled  = is_swizzled;
    texture->vram      = 0;
    texture->mipmaps   = lbound(num_levels,1) - 1;
    texture->palette   = palette;
    texture->antialias = 1;
    texture->repeat_u  = 1;
    texture->repeat_v  = 1;
    texture->lock_buf  = NULL;

    /* If texture data was given but we're not reusing the input buffer,
     * copy the texture data into the newly allocated buffer.  (Palettes
     * for 8-bit indexed data were copied above.) */

    if (!do_reuse && num_levels > 0) {
        int level_w = width;
        int level_h = height;
        int in_stride = stride;
        for (int level = 0; level < num_levels; level++) {
            if (is_aligned) {
                in_stride = align_up(in_stride, 128/bpp);
            }
            const int out_stride = align_up(level_w, 128/bpp);
            const uint8_t *pixels =
                (const uint8_t *)data + level_offsets[level];
            if (bpp == 8 && !is_alpha) {
                pixels += 256*4;  // Skip the palette data.
            }
            if (out_stride == in_stride) {
                memcpy(texture->pixels[level], pixels,
                       level_h * in_stride * (bpp/8));
            } else {
                const uint8_t *src = pixels;
                uint8_t *dest = texture->pixels[level];
                for (int y = 0; y < level_h;
                     y++, src += (in_stride*bpp)/8, dest += (out_stride*bpp)/8)
                {
                    memcpy(dest, src, (level_w * bpp + 7) / 8);
                }
            }
            level_w = lbound(level_w/2, 1);
            level_h = lbound(level_h/2, 1);
            in_stride = lbound(in_stride/2, 1);
        }
        sceKernelDcacheWritebackRange(texture->pixels[0],
                                      texture_limit - texture->pixels[0]);
    }

    /* All done! */

    if (reuse) {
        mem_free(data);
    }
    return texture;

  error_return:
    if (reuse) {
        mem_free(data);
    }
    return NULL;
}

/*-----------------------------------------------------------------------*/

void sys_texture_destroy(SysTexture *texture)
{
    PRECOND(texture != NULL, return);

    if (current_texture == texture) {
        ge_disable(GE_STATE_TEXTURE);
        current_texture = NULL;
    }

    /* We allocate textures as a single memory block, so we just need to
     * free that one block (plus the lock buffer if necessary).  However,
     * if we're in the middle of a frame, the GE might still need to draw
     * from this texture, so in that case we defer the final free operation
     * until the end of the frame.  Deferred destroy operations are managed
     * through a simple linked list, using the first word of the SysTexture
     * structure memory as the link pointer (since we don't need the
     * contents of the structure anymore). */
    mem_free(texture->lock_buf);
    if (psp_is_ge_busy()) {
        *(SysTexture **)texture = deferred_destroy_list;
        deferred_destroy_list = texture;
    } else {
        if (texture->vram) {
            psp_vram_free(texture->pixels[0]);
        }
        mem_free(texture);
    }
}

/*************************************************************************/
/*************** Interface: Texture information retrieval ****************/
/*************************************************************************/

int sys_texture_width(SysTexture *texture)
{
    return texture->width;
}

/*-----------------------------------------------------------------------*/

int sys_texture_height(SysTexture *texture)
{
    return texture->height;
}

/*-----------------------------------------------------------------------*/

int sys_texture_has_mipmaps(SysTexture *texture)
{
    return texture->mipmaps > 0;
}

/*************************************************************************/
/****************** Interface: Pixel data manipulation *******************/
/*************************************************************************/

int sys_texture_grab(SysTexture *texture, int x, int y, int w, int h,
                     UNUSED int readable)
{
    PRECOND(texture != NULL, return 0);
    PRECOND(w > 0 && w <= texture->width, return 0);
    PRECOND(h > 0 && h <= texture->height, return 0);

    if (UNLIKELY(x < 0)
     || UNLIKELY(y < 0)
     || UNLIKELY(x + w > DISPLAY_WIDTH)
     || UNLIKELY(y + h > DISPLAY_HEIGHT)) {
        DLOG("Grab rectangle outside screen bounds: (%d,%d)+%dx%d",
             x, y, w, h);
        return 0;
    }

    if (texture->format != GE_TEXFMT_8888) {
        DLOG("Can't copy into %dbpp texture",
             texture->format == GE_TEXFMT_T8 ? 8 : 16);
        return 0;
    }

    /* If we're only copying to part of the texture, we may need to
     * deswizzle the pixel data depending on the copy size. */
    if (texture->swizzled
        && ((w != texture->width  && w%4 != 0)
         || (h != texture->height && h%8 != 0)))
    {
        DLOG("Can't copy %dx%d into swizzled texture, deswizzling", w, h);
        /* For simplicity, we just reuse the lock/unlock logic, copying the
         * unswizzled lock buffer over the original pixel buffer. */
        uint8_t *rgba = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, texture->width, texture->height);
        if (UNLIKELY(!rgba)) {
            DLOG("Couldn't get RGBA buffer, aborting copy");
            return 0;
        }
        memcpy(texture->pixels[0], rgba,
               texture->stride * texture->height * 4);
        texture->swizzled = 0;
        sys_texture_unlock(texture, 0);
    }

    /* Do the actual copy. */
    fb_to_texture(x, y, w, h, texture);

    return 1;
}

/*-----------------------------------------------------------------------*/

uint8_t *sys_texture_lock(SysTexture *texture, SysTextureLockMode lock_mode,
                          int x, int y, int w, int h)
{
    PRECOND(texture != NULL, return NULL);
    PRECOND(x >= 0, return NULL);
    PRECOND(y >= 0, return NULL);
    PRECOND(w > 0, return NULL);
    PRECOND(h > 0, return NULL);
    PRECOND(x+w <= texture->width, return NULL);
    PRECOND(y+h <= texture->height, return NULL);

    if (texture->swizzled
     && ((texture->height & 7) != 0
         || (texture->format==GE_TEXFMT_T8 ? (texture->stride & 15) != 0 :
             texture->format==GE_TEXFMT_8888 ? (texture->stride & 3) != 0 :
             (texture->stride & 7) != 0)))
    {
        /* Safety check -- this combination is invalid anyway, but if we
         * try unswizzling with an unaligned height or stride, we'll
         * overrun the lock buffer. */
        DLOG("Can't handle swizzled textures with unaligned height or"
             " stride (%p, %dx%d, stride=%d)", texture, texture->width,
             texture->height, texture->stride);
        return NULL;
    }

    const uint32_t size = w * h * 4;
    texture->lock_buf = mem_alloc(size, 0, MEM_ALLOC_TEMP);
    if (UNLIKELY(!texture->lock_buf)) {
        DLOG("lock(%p): Failed to get a lock buffer (%u bytes)", texture, size);
        return NULL;
    }

    if (lock_mode == SYS_TEXTURE_LOCK_NORMAL) {

        if (w == texture->stride && h == texture->height) {
            texture_to_RGBA(texture, texture->lock_buf);

        } else if (!texture->swizzled) {
            uint32_t *dest = texture->lock_buf;
            const int stride = texture->stride;
            if (texture->format == GE_TEXFMT_T8) {
                const uint32_t * const palette = texture->palette;
                const uint8_t *src = texture->pixels[0] + y*stride + x;
                for (int yy = 0; yy < h; yy++, src += stride - w) {
                    for (int xx = 0; xx < w; xx++, src++, dest++) {
                        *dest = palette[*src];
                    }
                }
            } else if (texture->format == GE_TEXFMT_8888) {
                const uint32_t *src =
                    (const uint32_t *)((void *)texture->pixels[0])
                    + y*stride + x;
                for (int yy = 0; yy < h; yy++, src += stride, dest += w) {
                    memcpy(dest, src, w*4);
                }
            } else {  // 16bpp
                const int format = texture->format;
                const uint16_t *src =
                    (const uint16_t *)((void *)texture->pixels[0])
                    + y*stride + x;
                for (int yy = 0; yy < h; yy++, src += stride - w) {
                    for (int xx = 0; xx < w; xx++, src++, dest++) {
                        *dest = pixel_16bpp_to_RGBA(format, *src);
                    }
                }
            }

        } else {  // texture->swizzled
            /* For normal use this should be a rare case, so for simplicity,
             * we just deswizzle the entire texture and extract the desired
             * portion from the unswizzled data. */
            const uint32_t full_size = texture->stride * texture->height * 4;
            uint32_t *texture_buf = mem_alloc(full_size, 0, MEM_ALLOC_TEMP);
            if (UNLIKELY(!texture_buf)) {
                DLOG("lock(%p): Failed to get a texture image buffer (%u"
                     " bytes)", texture, full_size);
                mem_free(texture->lock_buf);
                texture->lock_buf = NULL;
                return NULL;
            }
            texture_to_RGBA(texture, texture_buf);
            const uint32_t *src = texture_buf + y*texture->width + x;
            uint32_t *dest = texture->lock_buf;
            if (w == texture->width) {
                memcpy(dest, src, h * texture->width * 4);
            } else {
                for (int yy = 0; yy < h; yy++, src += texture->stride, dest += w) {
                    memcpy(dest, src, w*4);
                }
            }
            mem_free(texture_buf);
        }

    }  // if (lock_mode == SYS_TEXTURE_LOCK_NORMAL) {

    texture->lock_x = x;
    texture->lock_y = y;
    texture->lock_w = w;
    texture->lock_h = h;
    return (uint8_t *)texture->lock_buf;
}

/*-----------------------------------------------------------------------*/

void sys_texture_unlock(SysTexture *texture, int update)
{
    PRECOND(texture != NULL, return);
    PRECOND(texture->lock_buf != NULL, return);

    if (update) {
        RGBA_to_texture(texture->lock_buf, texture->width,
                        texture->width, texture->height, 1, texture);
        sys_texture_flush(texture);
    }

    mem_free(texture->lock_buf);
    texture->lock_buf = NULL;
}

/*-----------------------------------------------------------------------*/

void sys_texture_flush(SysTexture *texture)
{
    PRECOND(texture != NULL, return);

    sceKernelDcacheWritebackRange(texture->pixels[0],
                                  texture->stride * texture->height * 4);
    if (current_texture == texture) {
        ge_flush_texture_cache();
    }
}

/*************************************************************************/
/******************** Interface: Rendering operations ********************/
/*************************************************************************/

void sys_texture_set_repeat(SysTexture *texture, int repeat_u, int repeat_v)
{
    texture->repeat_u = repeat_u;
    texture->repeat_v = repeat_v;
    if (current_texture == texture) {
        ge_set_texture_wrap_mode(
            texture->repeat_u ? GE_TEXWRAPMODE_REPEAT : GE_TEXWRAPMODE_CLAMP,
            texture->repeat_v ? GE_TEXWRAPMODE_REPEAT : GE_TEXWRAPMODE_CLAMP);
    }
}

/*-----------------------------------------------------------------------*/

void sys_texture_set_antialias(SysTexture *texture, int on)
{
    texture->antialias = on;
    if (current_texture == texture) {
        if (texture->antialias) {
            ge_set_texture_filter(
                GE_TEXFILTER_LINEAR, GE_TEXFILTER_LINEAR,
                texture->mipmaps ? GE_TEXMIPFILTER_LINEAR
                                 : GE_TEXMIPFILTER_NEAREST);
        } else {
            ge_set_texture_filter(
                GE_TEXFILTER_NEAREST, GE_TEXFILTER_NEAREST,
                GE_TEXMIPFILTER_NEAREST);
        }
    }
}

/*-----------------------------------------------------------------------*/

void sys_texture_apply(SysTexture *texture)
{
    if (texture != current_texture) {
        current_texture = texture;
        psp_set_texture_state();
    }
}

/*************************************************************************/
/******************** PSP-specific interface routines ********************/
/*************************************************************************/

void psp_texture_init(void)
{
    current_texture = NULL;
}

/*-----------------------------------------------------------------------*/

void psp_set_texture_state(void)
{
    if (current_texture) {

        SysTexture *texture = current_texture;  // For brevity.

        if (texture->format == GE_TEXFMT_T8) {
            ge_set_colortable(texture->palette, 256, GE_PIXFMT_8888, 0, 0xFF);
        }

        int width = texture->width, height = texture->height;
        int stride = texture->stride;
        const int format = texture->format;
        const int bpp = (format==GE_TEXFMT_T8 ? 8 :
                         format==GE_TEXFMT_8888 ? 32 : 16);
        for (unsigned int level = 0; level <= texture->mipmaps; level++) {
            ge_set_texture_data(level, texture->pixels[level],
                                width, height, stride);
            width  = lbound(width/2, 1);
            height = lbound(height/2, 1);
            stride = align_up(stride/2, 128/bpp);
        }

        ge_set_texture_format(texture->mipmaps + 1, texture->swizzled, format);
        ge_set_texture_draw_mode(GE_TEXDRAWMODE_MODULATE, 1);

        /* Use texture coordinate scaling to adjust texture coordinates
         * when the texture width or height is not a power of 2. */
        const int effective_w = ubound(texture->width, 512);
        const int effective_h = ubound(texture->height, 512);
        const int log2_width  = texture->width==1 ? 0 :
            ubound(32 - __builtin_clz(effective_w-1), 9);
        const int log2_height = texture->height==1 ? 0 :
            ubound(32 - __builtin_clz(effective_h-1), 9);
        ge_set_texture_scale((float)effective_w / (float)(1<<log2_width),
                             (float)effective_h / (float)(1<<log2_height));

        if (texture->antialias) {
            ge_set_texture_filter(GE_TEXFILTER_LINEAR, GE_TEXFILTER_LINEAR,
                                  texture->mipmaps ? GE_TEXMIPFILTER_LINEAR
                                                   : GE_TEXMIPFILTER_NEAREST);
        } else {
            ge_set_texture_filter(GE_TEXFILTER_NEAREST, GE_TEXFILTER_NEAREST,
                                  GE_TEXMIPFILTER_NEAREST);
        }

        ge_set_texture_wrap_mode(
            texture->repeat_u ? GE_TEXWRAPMODE_REPEAT : GE_TEXWRAPMODE_CLAMP,
            texture->repeat_v ? GE_TEXWRAPMODE_REPEAT : GE_TEXWRAPMODE_CLAMP);

        ge_enable(GE_STATE_TEXTURE);

    } else {  // current_texture == NULL

        ge_disable(GE_STATE_TEXTURE);

    }
}

/*-----------------------------------------------------------------------*/

SysTexture *psp_current_texture(void)
{
    return current_texture;
}

/*-----------------------------------------------------------------------*/

void psp_texture_flush_deferred_destroy_list(void)
{
    for (SysTexture *texture = deferred_destroy_list, *next;
         texture != NULL; texture = next)
    {
        next = *(SysTexture **)texture;
        if (texture->vram) {
            psp_vram_free(texture->pixels[0]);
        }
        mem_free(texture);
    }
    deferred_destroy_list = NULL;
}

/*-----------------------------------------------------------------------*/

SysTexture *psp_create_vram_texture(int width, int height)
{
    if (UNLIKELY(width <= 0) || UNLIKELY(height <= 0)) {
        DLOG("Invalid parameters: %d %d", width, height);
        return NULL;
    }

    const int stride = align_up(width, 4);

    void *pixels = psp_vram_alloc(stride * height * 4);
    if (UNLIKELY(!pixels)) {
        DLOG("No VRAM available for %dx%d pixel buffer", stride, height);
        return NULL;
    }

    SysTexture *texture = mem_alloc(sizeof(*texture), 0, MEM_ALLOC_TEMP);
    if (UNLIKELY(!texture)) {
        DLOG("Failed to allocate texture");
        psp_vram_free(pixels);
        return NULL;
    }
    mem_debug_set_info(texture, MEM_INFO_TEXTURE);

    texture->width     = width;
    texture->height    = height;
    texture->stride    = stride;
    texture->format    = GE_TEXFMT_8888;
    texture->swizzled  = 0;
    texture->vram      = 1;
    texture->mipmaps   = 0;
    texture->pixels[0] = pixels;
    texture->palette   = NULL;
    texture->antialias = 1;
    texture->repeat_u  = 1;
    texture->repeat_v  = 1;
    texture->lock_buf  = NULL;
    return texture;
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static void texture_to_RGBA(SysTexture *texture, uint32_t *dest)
{
    PRECOND(texture != NULL, return);
    PRECOND(dest != NULL, return);

    const int width = texture->width;
    const int height = texture->height;
    const int stride = texture->stride;
    const uint32_t * const palette = texture->palette;

    if (texture->swizzled) {

        if (texture->format == GE_TEXFMT_T8) {

            const uint8_t *src = texture->pixels[0];
            for (int y = 0; y < height; y += 8, dest += stride*7) {
                for (int x = 0; x < width; x += 16, dest += 16) {
                    uint32_t *destline = dest;
                    for (int line = 0; line < 8; line++, destline += width) {
                        for (int i = 0; i < 16; i += 4, src += 4) {
                            const uint8_t pixel0 = src[0];
                            const uint8_t pixel1 = src[1];
                            const uint8_t pixel2 = src[2];
                            const uint8_t pixel3 = src[3];
                            destline[i+0] = palette[pixel0];
                            destline[i+1] = palette[pixel1];
                            destline[i+2] = palette[pixel2];
                            destline[i+3] = palette[pixel3];
                        }
                    }
                }
            }

        } else if (texture->format == GE_TEXFMT_8888) {

            const uint32_t *src =
                (const uint32_t *)((void *)texture->pixels[0]);
            for (int y = 0; y < height; y += 8, dest += stride*7) {
                for (int x = 0; x < width; x += 4, dest += 4) {
                    uint32_t *destline = dest;
                    for (int line = 0; line < 8;
                         line++, src += 4, destline += stride)
                    {
                        const uint32_t pixel0 = src[0];
                        const uint32_t pixel1 = src[1];
                        const uint32_t pixel2 = src[2];
                        const uint32_t pixel3 = src[3];
                        destline[0] = pixel0;
                        destline[1] = pixel1;
                        destline[2] = pixel2;
                        destline[3] = pixel3;
                    }
                }
            }

        } else {  // 16bpp

            const int format = texture->format;
            const uint16_t *src =
                (const uint16_t *)((void *)texture->pixels[0]);
            for (int y = 0; y < height; y += 8, dest += stride*7) {
                for (int x = 0; x < width; x += 8, dest += 8) {
                    uint32_t *destline = dest;
                    for (int line = 0; line < 8; line++, destline += stride) {
                        for (int i = 0; i < 8; i++, src++) {
                            destline[i] = pixel_16bpp_to_RGBA(format, *src);
                        }
                    }
                }
            }

        }

    } else {  // !swizzled

        if (texture->format == GE_TEXFMT_T8) {
            const uint8_t *src = texture->pixels[0];
            for (int y = 0; y < height; y++, src += stride - width) {
                for (int x = 0; x < width; x++, src++, dest++) {
                    *dest = palette[*src];
                }
            }
        } else if (texture->format == GE_TEXFMT_8888) {
            const uint32_t *src =
                (const uint32_t *)((void *)texture->pixels[0]);
            for (int y = 0; y < height; y++, src += stride, dest += width) {
                memcpy(dest, src, width*4);
            }
        } else {  // 16bpp
            const int format = texture->format;
            const uint16_t *src =
                (const uint16_t *)((void *)texture->pixels[0]);
            for (int y = 0; y < height; y++, src += stride - width) {
                for (int x = 0; x < width; x++, src++, dest++) {
                    *dest = pixel_16bpp_to_RGBA(format, *src);
                }
            }
        }

    }  // if (swizzled)
}

/*-----------------------------------------------------------------------*/

static inline uint32_t pixel_16bpp_to_RGBA(
    const int format, const uint16_t pixel)
{
    if (format == GE_TEXFMT_5650) {
        const uint32_t b = pixel>>11 & 0x1F;
        const uint32_t g = pixel>> 5 & 0x3F;
        const uint32_t r = pixel>> 0 & 0x1F;
        return 0xFF000000
             | (b<<3 | b>>2) << 16
             | (g<<2 | g>>4) <<  8
             | (r<<3 | r>>2) <<  0;
    } else if (format == GE_TEXFMT_5551) {
        const uint32_t a = pixel>>15 & 0x01;
        const uint32_t b = pixel>>10 & 0x1F;
        const uint32_t g = pixel>> 5 & 0x1F;
        const uint32_t r = pixel>> 0 & 0x1F;
        return (a ? 0xFF000000 : 0)
             | (b<<3 | b>>2) << 16
             | (g<<3 | g>>2) <<  8
             | (r<<3 | r>>2) <<  0;
    } else {  // format == GE_TEXFMT_4444
        const uint32_t a = pixel>>12 & 0x0F;
        const uint32_t b = pixel>> 8 & 0x0F;
        const uint32_t g = pixel>> 4 & 0x0F;
        const uint32_t r = pixel>> 0 & 0x0F;
        return (a<<4 | a) << 24
             | (b<<4 | b) << 16
             | (g<<4 | g) <<  8
             | (r<<4 | r) <<  0;
    }
}

/*-----------------------------------------------------------------------*/

static void RGBA_to_texture(const uint32_t *src, int src_stride,
                            int width, int height, int has_alpha,
                            SysTexture *texture)
{
    PRECOND(src != NULL, return);
    PRECOND(texture != NULL, return);
    PRECOND(width <= texture->width, return);
    PRECOND(height <= texture->height, return);
    PRECOND(!texture->swizzled || (width%4 == 0 && height%8 == 0), return);

    const uint32_t alpha_mod = has_alpha ? 0 : 0xFF000000;
    uint32_t *dest = (uint32_t *)((void *)texture->pixels[0]);
    uint32_t dest_stride = texture->stride;

    if (texture->swizzled) {
        for (int y = 0; y < height;
             y += 8, src += src_stride*8, dest += dest_stride*8)
        {
            uint32_t *dest_row = dest;
            for (int x = 0; x < width; x += 4) {
                const uint32_t *src_line = &src[x];
                for (int line = 0; line < 8;
                     line++, src_line += src_stride, dest_row += 4)
                {
                    const uint32_t pixel0 = src_line[0];
                    const uint32_t pixel1 = src_line[1];
                    const uint32_t pixel2 = src_line[2];
                    const uint32_t pixel3 = src_line[3];
                    dest_row[0] = pixel0 | alpha_mod;
                    dest_row[1] = pixel1 | alpha_mod;
                    dest_row[2] = pixel2 | alpha_mod;
                    dest_row[3] = pixel3 | alpha_mod;
                }
            }
        }
    } else if (width % 4 == 0) {
        for (int y = 0; y < height;
             y++, src += src_stride - width, dest += dest_stride - width)
        {
            for (int x = 0; x < width; x += 4, src += 4, dest += 4) {
                const uint32_t pixel0 = src[0];
                const uint32_t pixel1 = src[1];
                const uint32_t pixel2 = src[2];
                const uint32_t pixel3 = src[3];
                dest[0] = pixel0 | alpha_mod;
                dest[1] = pixel1 | alpha_mod;
                dest[2] = pixel2 | alpha_mod;
                dest[3] = pixel3 | alpha_mod;
            }
        }
    } else {
        for (int y = 0; y < height;
             y++, src += src_stride - width, dest += dest_stride - width)
        {
            for (int x = 0; x < width; x++, src++, dest++) {
                *dest = *src | alpha_mod;
            }
        }
    }
}

/*-----------------------------------------------------------------------*/

static void fb_to_texture(int x0, int y0, int width, int height,
                          SysTexture *texture)
{
    PRECOND(x0 + width <= DISPLAY_WIDTH, return);
    PRECOND(y0 + height <= DISPLAY_HEIGHT, return);
    PRECOND(texture != NULL, return);
    PRECOND(width <= texture->stride, return);
    PRECOND(height <= texture->height, return);

    psp_sync_framebuffer(x0, y0, width, height);

    /* Tweak source parameters because we need to start from the bottom
     * rather than the top when writing to a texture. */
    const uint32_t *src =
        psp_fb_pixel_address(x0, (psp_framebuffer_height()-1) - y0);
    const int fb_stride = psp_framebuffer_stride();
    const int src_stride = -fb_stride;

    RGBA_to_texture(src, src_stride, width, height, 0, texture);
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
